import java.util.*;
import java.io.*;

/**
 * Class that contains information about different board layouts
 * Handles making jumps, moves and finding out what pieces are where.
 */
public class Position
{
   private boolean redsMove;//Determines who's move it is
   private boolean[] empty;//whether there's a piece or not
   private boolean[] black;//whether the piece is black or red
   private boolean[] piece; //whether a piece is a king or not
   
   //is it red's move or blacks
   public boolean isRedsMove() { return redsMove; }

   //is a space empty or not
   public boolean isEmpty(int square) { return empty[square - 1]; }

   //is the space black or red
   public boolean isBlack(int square) { return black[square - 1]; }

   //is the piece a King or not
   public boolean isPiece(int square) { return piece[square - 1]; }

   //makes copy of empty array
   public boolean[] getEmpty()  {return (boolean[])empty.clone();}
  
   //makes copy of black array
   public boolean[] getBlack()  {return (boolean[])black.clone();}
  
   //makes copy of piece array
   public boolean[] getPiece()  {return (boolean[])piece.clone();}
  
  //constructs initial board position
   public Position()
   {
      empty = new boolean[LetsPlay.SQUARES];
      black = new boolean[LetsPlay.SQUARES];
      piece = new boolean[LetsPlay.SQUARES];

      // loop through and setup empty positions, array 12-19
      for(int i=12; i<20; i++) 
        {empty[i] = true;}

      // loop through and setup the red positions, array 0-11
      for(int i=0; i<12; i++)
      {
         black[i] = false;
         piece[i] = true;
      }

      // loop through and setup the white positions, array 20-31
      for(int i=20; i<32; i++)
      {
         black[i] = true;
         piece[i] = true;
      }

      // set red's move to true
      redsMove = true;
   }

   //constructs a position from the OLD position
   //*basis for the redo function*
   public Position(Position oldPosition)
   {
      empty = oldPosition.getEmpty();
      black = oldPosition.getBlack();
      piece = oldPosition.getPiece();
      redsMove = oldPosition.isRedsMove();
   }

  //makes a new move
   public Position makeMove(String move)
   {
      // Check if move is illegal.
      if(!isLegalMove(move))
         throw new RuntimeException("Bad Move: " + move + "\n");

      // create a new postition from the current position
      Position newPosition = new Position(this);

      // create a string tokenizer for the moves
      StringTokenizer st = new StringTokenizer(move, "-");

      // create an array of ints for the moves
      int[] moves = new int[st.countTokens()];

      int index = 0;

      // populate the moves array
      while(st.hasMoreTokens())
         moves[index++] = Integer.parseInt(st.nextToken());

      int finalSquare = moves[moves.length-1];

      // loop through the moves looking for jumps
      for(int i=0; i<moves.length-1; i++)
      {
         int jumpedSquare = getJumpedSquare(moves[i], moves[i+1]);

         if (jumpedSquare > 0)
            newPosition.empty[jumpedSquare-1] = true;
      }

      // set new square to !empty and whatever kind of piece it was
      newPosition.empty[finalSquare-1] = false;
      newPosition.piece[finalSquare-1] = this.piece[moves[0]-1];
      newPosition.black[finalSquare-1] = this.black[moves[0]-1];

      // set old square to empty
      newPosition.empty[moves[0]-1] = true;

      // If in king row then not a piece.
      if(isKingRow(finalSquare))
         newPosition.piece[finalSquare-1] = false;

      newPosition.redsMove = ! isRedsMove();

      return newPosition;
   }

   //Gets the number of the square jumped.
   private static int getJumpedSquare(int startSquare, int endSquare)
   {
      // check in all 4 directions
      if(getUpperLeft(getUpperLeft(startSquare)) == endSquare)
         return getUpperLeft(startSquare);

      if(getUpperRight(getUpperRight(startSquare)) == endSquare)
         return getUpperRight(startSquare);

      if(getLowerLeft(getLowerLeft(startSquare)) == endSquare)
         return getLowerLeft(startSquare);

      if(getLowerRight(getLowerRight(startSquare)) == endSquare)
         return getLowerRight(startSquare);

      return -1;
   }

   //Checks input move to see if legal
   public boolean isLegalMove(String move)
   {
      boolean isLegal = false;

      // get the array list of legal moves
      ArrayList legalMoves = getLegalMoves();

      // check to see if this one is in there
      for(int i=0; i<legalMoves.size(); i++)
      {
         if(move.equals(legalMoves.get(i)))
            isLegal = true;
      }

      return isLegal;
   }

   //Returns an ArrayList of all currently legal moves
   public ArrayList getLegalMoves()
   {
      ArrayList result = new ArrayList();

      // boolean for must jump
      boolean mustJump = false;

      // first check if a jump must be made
      for(int i=0; i<LetsPlay.SQUARES; i++)
         if (! empty[i] && black[i] != redsMove)
            if (findJumps(result, String.valueOf(i + 1)))
               mustJump = true;

      if (mustJump) return result;

      // check whose move it is
      if(isRedsMove())
      {
         // now just check if moves for red pieces or kings can be made
         for(int i=0; i<LetsPlay.SQUARES; i++)
         {
            // check for red piece or king in this position
            if(!black[i]&&!empty[i])
            {
               // check for empty to lowerLeft, check if it exists first
               if(getLowerLeft(i+1) != 0)
               {
                  if(empty[getLowerLeft(i+1)-1])
                  {
                     result.add(new String((i+1) + "-" + (getLowerLeft(i+1))));
                  }
               }
               // check for empty to lowerRight, check if it exists first
               if(getLowerRight(i+1) != 0)
               {
                  if(empty[getLowerRight(i+1)-1])
                  {
                     result.add(new String((i+1) + "-" + (getLowerRight(i+1))));
                  }
               }
            }
            // now check for red king here, 'cause they can go up too
            if(!black[i]&&!empty[i]&&!piece[i])
            {
               // check for empty to upperLeft, check if it exists first
               if(getUpperLeft(i+1) != 0)
               {
                  if(empty[getUpperLeft(i+1)-1])
                  {
                     result.add(new String((i+1) + "-" + (getUpperLeft(i+1))));
                  }
               }
               // check for empty to upperRight, check if it exists first
               if(getUpperRight(i+1) != 0)
               {
                  if(empty[getUpperRight(i+1)-1])
                  {
                     result.add(new String((i+1) + "-" + (getUpperRight(i+1))));
                  }
               }
            }
         }
      }
      else
      {
         // now just check if moves for white pieces can be made
         for(int i=0; i<LetsPlay.SQUARES; i++)
         {
            // check for white piece or king in this position
            if(black[i]&&!empty[i])
            {
               // check for empty to upperLeft, check if it exists first
               if(getUpperLeft(i+1) != 0)
               {
                  if(empty[getUpperLeft(i+1)-1])
                  {
                     result.add(new String((i+1) + "-" + (getUpperLeft(i+1))));
                  }
               }
               // check for empty to upperRight, check if it exists first
               if(getUpperRight(i+1) != 0)
               {
                  if(empty[getUpperRight(i+1)-1])
                  {
                     result.add(new String((i+1) + "-" + (getUpperRight(i+1))));
                  }
               }
            }
            // now check for white king here, 'cause they can go down too
            if(black[i]&&!empty[i]&&!piece[i])
            {
               // check for empty to lowerLeft, check if it exists first
               if(getLowerLeft(i+1) != 0)
               {
                  if(empty[getLowerLeft(i+1)-1])
                  {
                     result.add(new String((i+1) + "-" + (getLowerLeft(i+1))));
                  }
               }
               // check for empty to lowerRight, check if it exists first
               if(getLowerRight(i+1) != 0)
               {
                  if(empty[getLowerRight(i+1)-1])
                  {
                     result.add(new String((i+1) + "-" + (getLowerRight(i+1))));
                  }
               }
            }
         }
      }

      return result;
   }

   //finds whether there are jumps or not
   private boolean findJumps(ArrayList result, String moveStart)
   {
      boolean isBlack = ! redsMove;
      boolean isPiece = false;
      boolean foundJumps = false;
      Position now = new Position(this);
      int jumpSquare = 0;
      int oldSquare = 0;

      //Update Current to reflect moveStart
      String split = new String(moveStart);
      StringTokenizer tokenizer = new StringTokenizer(split, "-");

      while (tokenizer.hasMoreTokens())
      {
         jumpSquare = Integer.parseInt(tokenizer.nextToken());
         now.empty[jumpSquare - 1] = true;

         if (oldSquare == 0)
            isPiece = isPiece(jumpSquare);
         else
            now.empty[getJumpedSquare(oldSquare, jumpSquare) - 1] = true;

         oldSquare = jumpSquare;
      }

      //Look for jumps in each direction from jumpSquare
      for (int direction = 0; direction < 4; direction++)
      {
         int jumped = 0;
         int target = 0;

         switch (direction)
         {
            case 0:
               jumped = (! isBlack && isPiece) ? 0 : getUpperLeft(jumpSquare);
               target = getUpperLeft(jumped);
               break;
            case 1:
               jumped = (! isBlack && isPiece) ? 0 : getUpperRight(jumpSquare);
               target = getUpperRight(jumped);
               break;
            case 2:
               jumped = (  isBlack && isPiece) ? 0 : getLowerLeft(jumpSquare);
               target = getLowerLeft(jumped);
               break;
            case 3:
               jumped = (  isBlack && isPiece) ? 0 : getLowerRight(jumpSquare);
               target = getLowerRight(jumped);
               break;
         }

         if (jumped != 0 &&         target != 0 &&
             ! now.isEmpty(jumped) && now.isBlack(jumped) != isBlack &&
             now.isEmpty(target))
         {
            foundJumps = true;
            String newMove = moveStart + "-" + String.valueOf(target);
            if (! findJumps(result, newMove))
               result.add(newMove);
         }

      }

      return foundJumps;
   }

   // the following cases are necessary when using a 1-D array
   public static int getUpperLeft(int square)
   {
      switch (square)
      {
         case  6: case  7: case  8:
         case 14: case 15: case 16:
         case 22: case 23: case 24:
         case 30: case 31: case 32:
            return square - 5;
         case  9: case 10: case 11: case 12:
         case 17: case 18: case 19: case 20:
         case 25: case 26: case 27: case 28:
            return square - 4;
         default:
            return 0;
      }
   }

   public static int getUpperRight(int square)
   {
      switch (square)
      {
         case  5: case  6: case  7: case  8:
         case 13: case 14: case 15: case 16:
         case 21: case 22: case 23: case 24:
         case 29: case 30: case 31: case 32:
            return square - 4;
         case  9: case 10: case 11:
         case 17: case 18: case 19:
         case 25: case 26: case 27:
            return square - 3;
         default: return 0;
      }
   }

   public static int getLowerLeft(int square)
   {
      switch (square)
      {
         case  1: case  2: case  3: case  4:
         case  9: case 10: case 11: case 12:
         case 17: case 18: case 19: case 20:
         case 25: case 26: case 27: case 28:
            return square + 4;
         case  6: case  7: case  8:
         case 14: case 15: case 16:
         case 22: case 23: case 24:
            return square + 3;
         default: return 0;
      }
   }


   public static int getLowerRight(int square)
   {
      switch (square)
      {
         case  1: case  2: case  3:
         case  9: case 10: case 11:
         case 17: case 18: case 19:
         case 25: case 26: case 27:
            return square + 5;
         case  5: case  6: case  7: case  8:
         case 13: case 14: case 15: case 16:
         case 21: case 22: case 23: case 24:
            return square + 4;
         default: return 0;
      }
   }

   //Returns the number of red pieces excluding kings in board position.
   public int countRedPieces()
   {
      // acumulator for red pieces
      int numOfRedPieces = 0;

      // count the number of red pieces that are not kings
      for(int i=0; i<LetsPlay.SQUARES; i++)
      {
         // red piece when (white == false && piece == true && empty == false)
         if(!black[i]&&piece[i]&&!empty[i])
            numOfRedPieces++;
      }
      return numOfRedPieces;
   }

   //Returns the number of white pieces excluding kings in board position.
   public int countBlackPieces()
   {
      // acumulator for white pieces
      int numOfBlackPieces = 0;

      // count the number of white pieces
      for(int i=0; i<LetsPlay.SQUARES; i++)
      {
         // black piece when (black == true && piece == true && empty == false)
         if(black[i]&&piece[i]&&!empty[i])
            numOfBlackPieces++;
      }
      return numOfBlackPieces;
   }


   //Returns the number of red kings in board position.
   public int countRedKings()
   {
      // acumulator for number of red kings
      int numOfRedKings = 0;

      // count the number of red kings
      for(int i=0; i<LetsPlay.SQUARES; i++)
      {
         // red king when (white == false && piece == false && empty == false)
         if(!black[i]&&!piece[i]&&!empty[i])
            numOfRedKings++;
      }
      return numOfRedKings;
   }

   /**
    * Returns the number of white kings in board position.
    */
   public int countBlackKings()
   {
      // acumulator for number of white kings
      int numOfBlackKings = 0;

      // count the number of white kings
      for(int i=0; i<LetsPlay.SQUARES; i++)
      {
         // white king when (white == true && piece == false && empty == false)
         if(black[i]&&!piece[i]&&!empty[i])
            numOfBlackKings++;
      }
      return numOfBlackKings;
   }

   /**
    * Is the square in the king row for the current player?
    */
   public boolean isKingRow(int square)
   {
      return ( (redsMove && square > 28) || (!redsMove && square < 5) );
   }
}
